use std::{error::Error, sync::Arc};

use vulkano::{
    Validated,
    buffer::{AllocateBufferError, Buffer, BufferContents, BufferCreateInfo, Subbuffer},
    memory::allocator::{AllocationCreateInfo, StandardMemoryAllocator},
};

pub trait AsBindableBuffer<T> {
    type BufferData: BufferContents + Clone;

    fn buffer_create_info() -> BufferCreateInfo;

    fn allocation_create_info() -> AllocationCreateInfo;

    fn to_buffer_data(data: &T) -> Self::BufferData;

    fn create_buffer(
        memory_allocator: &Arc<StandardMemoryAllocator>,
        data: &T,
    ) -> Result<Subbuffer<[Self::BufferData]>, Validated<AllocateBufferError>> {
        let buffer_data = Self::to_buffer_data(data);
        Buffer::from_iter(
            memory_allocator.clone(),
            Self::buffer_create_info(),
            Self::allocation_create_info(),
            [buffer_data],
        )
    }

    fn update_buffer(
        buffer: &Subbuffer<[Self::BufferData]>,
        data: &T,
    ) -> Result<(), Box<dyn Error>> {
        let buffer_data = Self::to_buffer_data(data);
        let mut write_guard = buffer.write()?;
        write_guard[0] = buffer_data;
        Ok(())
    }
}

pub trait AsUniformBuffer<T>: AsBindableBuffer<T> {
    fn create_uniform_buffer(
        memory_allocator: &Arc<StandardMemoryAllocator>,
        data: &T,
    ) -> Result<Subbuffer<[Self::BufferData]>, Validated<AllocateBufferError>> {
        Self::create_buffer(memory_allocator, data)
    }
}

pub trait AsVertexBuffer<T>: AsBindableBuffer<T> {
    fn create_vertex_buffer(
        memory_allocator: &Arc<StandardMemoryAllocator>,
        vertices: &[T],
    ) -> Result<Subbuffer<[Self::BufferData]>, Validated<AllocateBufferError>> {
        let buffer_data: Vec<Self::BufferData> =
            vertices.iter().map(|v| Self::to_buffer_data(v)).collect();

        Buffer::from_iter(
            memory_allocator.clone(),
            Self::buffer_create_info(),
            Self::allocation_create_info(),
            buffer_data,
        )
    }
}

pub trait AsIndexBuffer<T>: AsBindableBuffer<T> {
    fn create_index_buffer(
        memory_allocator: &Arc<StandardMemoryAllocator>,
        indices: &[T],
    ) -> Result<Subbuffer<[Self::BufferData]>, Validated<AllocateBufferError>> {
        let buffer_data: Vec<Self::BufferData> =
            indices.iter().map(|i| Self::to_buffer_data(i)).collect();

        Buffer::from_iter(
            memory_allocator.clone(),
            Self::buffer_create_info(),
            Self::allocation_create_info(),
            buffer_data,
        )
    }
}
