use std::collections::BTreeMap;
use std::sync::Arc;

use vulkano::buffer::{
    AllocateBufferError, BufferContents, BufferCreateInfo, BufferUsage, Subbuffer,
};
use vulkano::descriptor_set::allocator::{DescriptorSetAllocator, StandardDescriptorSetAllocator};
use vulkano::descriptor_set::layout::{
    DescriptorSetLayout, DescriptorSetLayoutBinding, DescriptorType,
};
use vulkano::descriptor_set::{DescriptorSet, WriteDescriptorSet};
use vulkano::memory::allocator::{AllocationCreateInfo, MemoryTypeFilter, StandardMemoryAllocator};
use vulkano::shader::ShaderStages;
use vulkano::{Validated, VulkanError};

use crate::core::render::primitives::{AsBindableBuffer, AsBindableDescriptorSet, AsUniformBuffer};

#[derive(BufferContents, Clone, Copy)]
#[repr(C)]
pub struct Mvp {
    pub model: [[f32; 4]; 4],
    pub view: [[f32; 4]; 4],
    pub projection: [[f32; 4]; 4],
}

impl Mvp {
    pub fn into_buffer(
        self,
        memory_allocator: &Arc<StandardMemoryAllocator>,
    ) -> Result<Subbuffer<[Mvp]>, Validated<AllocateBufferError>> {
        Self::create_uniform_buffer(memory_allocator, &self)
    }
}

impl AsBindableBuffer<Mvp> for Mvp {
    type BufferData = Mvp;

    fn buffer_create_info() -> BufferCreateInfo {
        BufferCreateInfo {
            usage: BufferUsage::UNIFORM_BUFFER,
            ..Default::default()
        }
    }

    fn allocation_create_info() -> AllocationCreateInfo {
        AllocationCreateInfo {
            memory_type_filter: MemoryTypeFilter::PREFER_DEVICE
                | MemoryTypeFilter::HOST_SEQUENTIAL_WRITE,
            ..Default::default()
        }
    }

    fn to_buffer_data(data: &Mvp) -> Self::BufferData {
        *data
    }
}

impl AsUniformBuffer<Mvp> for Mvp {}

impl AsBindableDescriptorSet<Subbuffer<[Mvp]>> for Mvp {
    fn as_descriptor_set_layout_bindings() -> BTreeMap<u32, DescriptorSetLayoutBinding> {
        BTreeMap::<u32, DescriptorSetLayoutBinding>::from_iter([(
            0,
            DescriptorSetLayoutBinding {
                stages: ShaderStages::VERTEX,
                ..DescriptorSetLayoutBinding::descriptor_type(DescriptorType::UniformBuffer)
            },
        )])
    }

    fn as_descriptor_set(
        descriptor_set_allocator: &Arc<StandardDescriptorSetAllocator>,
        layout: &Arc<DescriptorSetLayout>,
        data: &Subbuffer<[Mvp]>,
    ) -> Result<Arc<DescriptorSet>, Validated<VulkanError>> {
        DescriptorSet::new(
            descriptor_set_allocator.clone(),
            layout.clone(),
            [WriteDescriptorSet::buffer(0, data.clone())],
            [],
        )
    }
}
