use std::{
    collections::HashMap,
    error::Error,
    fmt::{self, Debug, Display},
    hash::Hash,
    sync::Arc,
};

#[derive(Debug, Clone)]
pub struct ResourceLoadError {
    pub message: String,
}

impl Display for ResourceLoadError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "Resource load error: {}", self.message)
    }
}

impl Error for ResourceLoadError {}

impl ResourceLoadError {
    pub fn new(message: impl Into<String>) -> Self {
        Self {
            message: message.into(),
        }
    }
}

pub trait AsResourceManager<THandle, TResource>
where
    THandle: Clone + Eq + Hash + Debug,
{
    /// Type d'erreur lors du chargement
    type LoadError: Error;

    /// Charge une ressource avec l'handle donné
    fn load(&mut self, handle: THandle, resource: TResource) -> Result<(), Self::LoadError>;

    /// Récupère une ressource par son handle
    fn get(&self, handle: &THandle) -> Option<&TResource>;

    /// Récupère une ressource mutable par son handle
    fn get_mut(&mut self, handle: &THandle) -> Option<&mut TResource>;

    /// Supprime une ressource
    fn unload(&mut self, handle: &THandle) -> Option<TResource>;

    /// Vérifie si une ressource est chargée
    fn is_loaded(&self, handle: &THandle) -> bool;

    /// Retourne tous les handles chargés
    fn loaded_handles(&self) -> Vec<THandle>;

    /// Nettoie toutes les ressources
    fn clear(&mut self);
}

/// Implémentation basique d'un gestionnaire de ressources
pub struct BasicResourceManager<THandle, TResource>
where
    THandle: Clone + Eq + Hash + Debug,
{
    resources: HashMap<THandle, TResource>,
}

impl<THandle, TResource> BasicResourceManager<THandle, TResource>
where
    THandle: Clone + Eq + Hash + Debug,
{
    pub fn new() -> Self {
        Self {
            resources: HashMap::new(),
        }
    }

    pub fn with_capacity(capacity: usize) -> Self {
        Self {
            resources: HashMap::with_capacity(capacity),
        }
    }
}

impl<THandle, TResource> Default for BasicResourceManager<THandle, TResource>
where
    THandle: Clone + Eq + Hash + Debug,
{
    fn default() -> Self {
        Self::new()
    }
}

impl<THandle, TResource> AsResourceManager<THandle, TResource>
    for BasicResourceManager<THandle, TResource>
where
    THandle: Clone + Eq + Hash + Debug,
{
    type LoadError = ResourceLoadError;

    fn load(&mut self, handle: THandle, resource: TResource) -> Result<(), Self::LoadError> {
        self.resources.insert(handle, resource);
        Ok(())
    }

    fn get(&self, handle: &THandle) -> Option<&TResource> {
        self.resources.get(handle)
    }

    fn get_mut(&mut self, handle: &THandle) -> Option<&mut TResource> {
        self.resources.get_mut(handle)
    }

    fn unload(&mut self, handle: &THandle) -> Option<TResource> {
        self.resources.remove(handle)
    }

    fn is_loaded(&self, handle: &THandle) -> bool {
        self.resources.contains_key(handle)
    }

    fn loaded_handles(&self) -> Vec<THandle> {
        self.resources.keys().cloned().collect()
    }

    fn clear(&mut self) {
        self.resources.clear();
    }
}

pub struct ThreadSafeResourceManager<THandle, TResource>
where
    THandle: Clone + Eq + Hash + Debug,
{
    resources: Arc<std::sync::RwLock<HashMap<THandle, Arc<TResource>>>>,
}

impl<THandle, TResource> ThreadSafeResourceManager<THandle, TResource>
where
    THandle: Clone + Eq + Hash + Debug,
{
    pub fn new() -> Self {
        Self {
            resources: Arc::new(std::sync::RwLock::new(HashMap::new())),
        }
    }

    pub fn load(&self, handle: THandle, resource: TResource) -> Result<(), ResourceLoadError> {
        let mut resources = self
            .resources
            .write()
            .map_err(|_| ResourceLoadError::new("Failed to acquire write lock"))?;
        resources.insert(handle, Arc::new(resource));
        Ok(())
    }

    pub fn get(&self, handle: &THandle) -> Option<Arc<TResource>> {
        let resources = self.resources.read().ok()?;
        resources.get(handle).cloned()
    }

    pub fn unload(&self, handle: &THandle) -> Option<Arc<TResource>> {
        let mut resources = self.resources.write().ok()?;
        resources.remove(handle)
    }

    pub fn is_loaded(&self, handle: &THandle) -> bool {
        if let Ok(resources) = self.resources.read() {
            resources.contains_key(handle)
        } else {
            false
        }
    }

    pub fn loaded_handles(&self) -> Vec<THandle> {
        if let Ok(resources) = self.resources.read() {
            resources.keys().cloned().collect()
        } else {
            Vec::new()
        }
    }

    pub fn clear(&self) {
        if let Ok(mut resources) = self.resources.write() {
            resources.clear();
        }
    }
}

impl<THandle, TResource> Clone for ThreadSafeResourceManager<THandle, TResource>
where
    THandle: Clone + Eq + Hash + Debug,
{
    fn clone(&self) -> Self {
        Self {
            resources: Arc::clone(&self.resources),
        }
    }
}

pub trait AsAsyncLoadable<THandle>
where
    THandle: Clone + Eq + Hash + Debug + Send + Sync,
{
    type Resource: Send + Sync;
    type LoadError: Error + Send + Sync;

    async fn load_async(handle: THandle) -> Result<Self::Resource, Self::LoadError>;
}
